/*
 * This file is part of the DITA Open Toolkit project hosted on
 * Sourceforge.net. See the accompanying license.txt file for 
 * applicable licenses.
 */

/*
 * (c) Copyright IBM Corp. 2004, 2005 All Rights Reserved.
 */
package org.dita.dost.reader;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import org.dita.dost.log.DITAOTJavaLogger;
import org.dita.dost.log.MessageUtils;
import org.dita.dost.module.Content;
import org.dita.dost.module.ContentImpl;
import org.dita.dost.util.Constants;
import org.dita.dost.util.StringUtils;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * DitaValReader reads and parses the information from ditaval file which
 * contains the information of filtering and flagging.
 * 
 * @author Zhang, Yuan Peng
 */
public class DitaValReader extends AbstractXMLReader {
	private HashMap filterMap;

	private ContentImpl content;

	private XMLReader reader;

	private DITAOTJavaLogger logger;

	private List imageList = null;

	private String ditaVal = null;

	/**
	 * Default constructor of DitaValReader class.
	 */
	public DitaValReader() {
		super();
		filterMap = new HashMap();
		content = null;
		logger = new DITAOTJavaLogger();
		imageList = new ArrayList(Constants.INT_256);

		try {
			if (System.getProperty(Constants.SAX_DRIVER_PROPERTY) == null) {
				// The default sax driver is set to xerces's sax driver
				StringUtils.initSaxDriver();
			}
			reader = XMLReaderFactory.createXMLReader();
			reader.setContentHandler(this);
		} catch (Exception e) {
			logger.logException(e);
		}

	}

	/**
	 * @see org.dita.dost.reader.AbstractReader#read(java.lang.String)
	 */
	public void read(String input) {
		ditaVal = input;

		try {
			reader.parse(new InputSource(new FileInputStream(input)));
		} catch (Exception e) {
			logger.logException(e);
		}
	}

	/**
	 * @see org.dita.dost.reader.AbstractReader#getContent()
	 */
	public Content getContent() {
		content = new ContentImpl();
		content.setCollection(filterMap.entrySet());
		return content;
	}

	/**
	 * @see org.xml.sax.ContentHandler#startElement(java.lang.String,
	 *      java.lang.String, java.lang.String, org.xml.sax.Attributes)
	 * 
	 */
	public void startElement(String uri, String localName, String qName,
			Attributes atts) throws SAXException {
		String flagImage = null;
		if(atts.getValue(Constants.ATTRIBUTE_NAME_IMG)!=null){
			flagImage = atts.getValue(Constants.ATTRIBUTE_NAME_IMG);
		}else if(atts.getValue(Constants.ATTRIBUTE_NAME_IMAGEREF)!=null){
			flagImage = atts.getValue(Constants.ATTRIBUTE_NAME_IMAGEREF);
		}

		if (Constants.ELEMENT_NAME_PROP.equals(qName)) {
			String action = atts.getValue(Constants.ELEMENT_NAME_ACTION);
			String attName = atts.getValue(Constants.ATTRIBUTE_NAME_ATT);
			String attValue = atts.getValue(Constants.ATTRIBUTE_NAME_VAL);
			//first to check if the att attribute and val attribute are null 
			//which is a default action for elements without mapping with the other filter val
			String key=null;
			if(attName==null){
				key=Constants.DEFAULT_ACTION;
			}else 
				if(attValue==null){
					key=attName;//default action for the specified attribute
				}
				else{
					key = attName + Constants.EQUAL + attValue;
				}

			if (action != null) {
				insertAction(action, key);
			}
		}

		/*
		 * Parse image files for flagging
		 */
		if (flagImage != null && !"".equals(flagImage.trim())) {
			String filterDir;
			if (new File(flagImage).isAbsolute()) {
				imageList.add(flagImage);
				return;
			}

			// img is a relative path to the .ditaval file
			filterDir = new File(new File(ditaVal).getAbsolutePath())
					.getParent();
			imageList.add(new File(filterDir, flagImage).getAbsolutePath());
		}
	}

	/**
	 * Insert action into filetermap if key not present in the map
	 * @param action
	 * @param key
	 */
	private void insertAction(String action, String key) {
		if (filterMap.get(key) == null) {
			filterMap.put(key, action);
		} else {
			Properties prop = new Properties();
			prop.put("%1", key);
			logger.logError(MessageUtils.getMessage("DOTJ007E", prop)
					.toString());
		}
	}

	/**
	 * Return the image list
	 * @return
	 */
	public List getImageList() {
		return imageList;
	}
	
	/**
	 * Return the filter map
	 * @return
	 */
	public HashMap getFilterMap() {
		return filterMap;
	}
}
